import logging
# import inspect

from django.contrib import auth

from ...permissions import is_authenticated
from ...utils import get_session_key, check_user_can_login
from ... import app_settings
from ...apps.groups.utils import set_default_user_groups, set_default_profile_groups

from .base import DjangoSsoAppAuthenticationBaseMiddleware

logger = logging.getLogger('django_sso_app.core.middleware')


class DjangoSsoAppAuthenticationAppMiddleware(DjangoSsoAppAuthenticationBaseMiddleware):

    def app_process_request(self, request, request_jwt, decoded_jwt):
        user = getattr(request, 'user', None)

        if not is_authenticated(user):
            # We are seeing this user for the first time in this session, attempt
            # to authenticate the user.
            if app_settings.APIGATEWAY_ENABLED:
                consumer_custom_id = get_session_key(request, '__dssoa__apigateway__consumer_custom_id')

                user = auth.authenticate(request=request,
                                         consumer_custom_id=consumer_custom_id,
                                         encoded_jwt=request_jwt,
                                         decoded_jwt=decoded_jwt)
            else:
                user = auth.authenticate(request=request,
                                         encoded_jwt=request_jwt,
                                         decoded_jwt=decoded_jwt)

            # set request user
            setattr(request, 'user', user)  # !!

        # further checks
        if is_authenticated(user):
            if app_settings.REPLICATE_PROFILE:
                check_user_can_login(user)
            else:
                if app_settings.SERVICE_SUBSCRIPTION_REQUIRED:
                    check_user_can_login(user, skip_profile_completion_checks=True)
                else:
                    check_user_can_login(user, skip_profile_completion_checks=True,
                                         skip_service_subscription_checks=True)

            if app_settings.MANAGE_USER_GROUPS:
                # default groups check
                set_default_user_groups(user)
                # update_profile_groups(user.sso_app_profile)  # profile groups are managed by backend

            logger.info('User "{}" authenticated successfully!'.format(user))



import logging

from django.dispatch import receiver
from django.contrib.auth.signals import user_logged_in, user_logged_out
from django.db.models.signals import pre_save, post_save

from ....utils import set_session_key
from ....tokens.utils import jwt_encode
from ....functions import get_random_string
from ....permissions import is_django_staff
from ...profiles.models import Profile

from .... import app_settings

from ..models import Device
from ..utils import get_session_key, get_or_create_request_device, remove_all_profile_devices, remove_profile_device

logger = logging.getLogger('django_sso_app.core.apps.devices.signals')


# profile

@receiver(post_save, sender=Profile)
def post_profile_saved(sender, instance, created, **kwargs):
    if kwargs['raw']:
        # https://github.com/django/django/commit/18a2fb19074ce6789639b62710c279a711dabf97
        return

    profile = instance
    user = instance.user

    if not is_django_staff(user):
        if not created:  # if instance.pk:
            logger.debug('Profile "{}" post_save signal'.format(profile))

            rev_updated = getattr(profile, '__rev_updated', False)

            if rev_updated:
                logger.info('Rev updated, removing all user devices for Profile "{}"'.format(profile))

                remove_all_profile_devices(profile)


# device

@receiver(pre_save, sender=Device)
def pre_device_save(sender, instance, **kwargs):
    if not app_settings.APIGATEWAY_ENABLED:
        user = instance.profile.user
        device = instance

        if not is_django_staff(user):
            logger.debug('Skip creating api gateway JWT for Device "{}"'.format(device))

            device.apigw_jwt_id = None
            device.apigw_jwt_key = None
            device.apigw_jwt_secret = device.apigw_jwt_secret or get_random_string(32)


# login

@receiver(user_logged_in)
def create_profile_device_on_login(sender, request, user, **kwargs):
    """
    Sent when a user logs in successfully.
    Assign device and JWT token to request object.

    Arguments sent with this signal:

    sender
        The class of the user that just logged in.
    request
        The current HttpRequest instance.
    user
        The user instance that just logged in.

    """

    logger.debug('devices user_logged_in signal for user "{}"'.format(user))

    set_session_key(request, '__dssoa__logged_in', True)

    if not is_django_staff(user):
        device = get_or_create_request_device(request)
        token = jwt_encode(device.get_jwt_payload(), device.apigw_jwt_secret)

        set_session_key(request, '__dssoa__device', device)
        set_session_key(request, '__dssoa__jwt_token', token)

        logger.info('User Logged in, request has device "{}"'.format(device))


# logout

@receiver(user_logged_out)
def delete_profile_devices_on_logout(sender, request, user, **kwargs):
    """
    Sent when a user logs out successfully.

    Arguments sent with this signal:


Sent when the logout method is called.

    sender
        As above: the class of the user that just logged out or ``None``
        if the user was not authenticated.

    request
        The current :class:`~django.http.HttpRequest` instance.

    user
        The user instance that just logged out or ``None`` if the
        user was not authenticated.

    """
    logger.debug('devices user_logged_out signal for "{}"'.format(user))

    if user is not None:
        set_session_key(request, '__dssoa__logged_out', True)

        request_device_fingerprint = get_session_key(request, '__dssoa__device__fingerprint', None)

        if app_settings.LOGOUT_DELETES_ALL_PROFILE_DEVICES:
            deleted_devices = remove_all_profile_devices(user.get_sso_app_profile())
        else:
            if request_device_fingerprint is not None:
                request_device = Device.objects.filter(profile=user.get_sso_app_profile(),
                                                       fingerprint=request_device_fingerprint).first()
                if request_device is not None:
                    deleted_devices = remove_profile_device(request_device)
                else:
                    logger.warning('Request device not found')
                    deleted_devices = 0
            else:
                deleted_devices = 0

        logger.info('({}) devices deleted for user "{}"'.format(deleted_devices, user))



from django.contrib.auth import get_user_model

from django_filters import rest_framework as filters

from .models import Profile


class ProfileFilter(filters.FilterSet):
    class Meta:
        model = Profile
        fields = {
            'user__username': ['exact'],
            'sso_id': ['exact'],
            'groups__name': ['exact'],
        }



import logging

from django.db.models.signals import m2m_changed, post_save
from django.dispatch import receiver
from django.contrib.auth import get_user_model

from ....permissions import is_django_staff
from ...profiles.models import Profile
from ..models import Group
from ..utils import set_default_user_groups, set_default_profile_groups

logger = logging.getLogger('django_sso_app.core.apps.groups.signals')
User = get_user_model()


@receiver(m2m_changed, sender=Profile.groups.through)
def signal_handler_when_user_profile_is_added_or_removed_from_group(action, instance, reverse, model, pk_set, using, **kwargs):
    """
    Update sso_rev when user enter/exit groups
    """

    profile = instance
    user = profile.user

    is_loaddata = getattr(user, '__dssoa__loaddata', False)
    is_creating = getattr(user, '__dssoa__creating', False)
    #is_apigateway_update = getattr(user, '__dssoa__apigateway_update', False)
    is_updating_groups = getattr(user, '__dssoa__updating_default_groups', False)

    must_update_rev = (not is_loaddata) and \
                      (not is_creating) and \
                      (not is_updating_groups) and \
                      (not is_django_staff(user))
    #                       (not is_apigateway_update) and \
    groups_updated = False

    if action == 'pre_add':
        groups_updated = True
        for pk in pk_set:
            group = Group.objects.get(id=pk)
            logger.info('Profile "{}" entered group "{}"'.format(profile, group))

    elif action == 'pre_remove':
        groups_updated = True
        for pk in pk_set:
            group = Group.objects.get(id=pk)
            logger.info('Profile "{}" exited from group "{}"'.format(profile, group))

    # updating rev

    if groups_updated and must_update_rev:
        profile.update_rev(True)


@receiver(post_save, sender=Profile)
def update_groups_on_user_profile_updated(sender, instance, created, **kwargs):
    """
    Profile model has been updated, check user profile completeness and default groups
    """
    if kwargs['raw']:
        # https://github.com/django/django/commit/18a2fb19074ce6789639b62710c279a711dabf97
        return

    profile = instance
    user = profile.user

    if created and not is_django_staff(user):
        # default groups
        set_default_profile_groups(profile)

        # check profile groups
        user_object_profile = getattr(profile.user, '__dssoa__profile__object', None)

        if user_object_profile is not None:
            group_names = user_object_profile.get('groups', [])
        else:
            group_names = []

        for gn in group_names:
            profile.add_to_group(gn, creating=True)


@receiver(post_save, sender=User)
def add_user_model_to_default_groups_on_creation(sender, instance, created, **kwargs):
    """
    User model has been updated
    """
    user = instance

    if kwargs['raw']:
        # https://github.com/django/django/commit/18a2fb19074ce6789639b62710c279a711dabf97
        return

    if created and not is_django_staff(user):
        set_default_user_groups(user)




import logging

from .... import app_settings
from . import elastic

logger = logging.getLogger('django_sso_app.core.apps.events')
ELASTICSEARCH_INDEX_URL = app_settings.EVENTS_HOST + '/{}/_search'.format(app_settings.EVENTS_INDEX)


def fetch_event_type_events(event_type: str, from_date: str, jwt: str = None, apikey: str = None) -> (object, int):
    """
    Returns event objects given event_type

    :param event_type:
    :param from_date:
    :param jwt:
    :param apikey:
    :return:
    """
    ES_QUERY = """{
      "from": 0,
      "size": 1000,
      "sort" : [
        { "doc.timestamp" : {"order" : "asc"}}
      ],
      "query": {
        "bool": {
          "filter": [
             {
              "range": {
                "doc.timestamp": {
                  "gte": "{{ from_date }}"
                }
              }
            },
            {
              "terms": {
                "doc.type": [
                  "{{ event_type }}"
                ]
              }
            }
          ]
        }
      }
    }"""

    def _fetch_data(event_type, from_date):
        es_url = ELASTICSEARCH_INDEX_URL
        es_query = elastic.get_es_query(ES_QUERY,
                                        event_type=event_type,
                                        from_date=from_date)
        # print('CALLING ES', es_query, es_url)
        return elastic.call_backend(es_url, es_query, jwt=jwt, apikey=apikey)

    def _cleaned_data(data):
        for d in data:
            yield d['_source']['doc']

    es_response = _fetch_data(event_type, from_date)

    instance_tasks, count = es_response['hits']['hits'], es_response['hits']['total']['value']

    # return generator and total hits count
    return _cleaned_data(instance_tasks), count


def sort_event_type_events(events, event_type):
    # filter related procedure events, sort ascending by timestamp and return
    return sorted(filter(lambda x: x['type'] == event_type, events),
                  key=lambda x: x['timestamp'], reverse=False)



from django.conf import settings

from django_sso_app import dist_name, __version__


def get_stats_info(_request):
    return {
        "APP_NAME": dist_name,
        "DEPLOYMENT_ENV": settings.DEPLOYMENT_ENV,
        "REPOSITORY_REV": settings.REPOSITORY_REV,
        "SEMANTIC_VER": __version__
    }


def settings_context(_request):
    return {"settings": settings}



import logging

from django.contrib.auth.models import AbstractUser
from django.utils.translation import gettext_lazy as _
from django.urls import reverse

from ...core.apps.users.models import DjangoSsoAppUserModelMixin

logger = logging.getLogger('django_sso_app.backend.users')


class User(AbstractUser, DjangoSsoAppUserModelMixin):

    class Meta:
        verbose_name = _('user')
        verbose_name_plural = _('users')

    def get_relative_url(self):
        return reverse("users:detail", kwargs={"username": self.username})



import logging

import base64
from io import BytesIO
from PIL import Image

from django import forms
from django.utils.translation import ugettext_lazy as _
from django.core.files.uploadedfile import InMemoryUploadedFile, TemporaryUploadedFile

from django_countries.fields import CountryField

from ... import app_settings
from .models import Profile

logger = logging.getLogger('django_sso_app.core.apps.profiles')


class ProfileForm(forms.Form):
    # role = forms.ChoiceField(
    #     label=_('Role'),
    #     choices=app_settings.PROFILE_ROLE_CHOICES,
    #     initial=0)

    ssn = forms.CharField(
        label=_('Social Security Number'),
        widget=forms.TextInput(
            attrs={'type': 'text',
                   'placeholder': _('Social Security Number')}))

    phone = forms.CharField(
        label=_('Phone'),
        widget=forms.TextInput(
            attrs={'type': 'text',
                   'placeholder': _('Phone')}))

    first_name = forms.CharField(
        label=_('First name'),
        widget=forms.TextInput(
            attrs={'type': 'text',
                   'placeholder': _('First name')}))

    last_name = forms.CharField(
        label=_('Last name'),
        widget=forms.TextInput(
            attrs={'type': 'text',
                   'placeholder': _('Last name')}))

    description = forms.CharField(
        label=_('Description'),
        widget=forms.Textarea(
            attrs={'placeholder': _('Description'),
                   'rows': 2}))

    picture = forms.ImageField(
        label=_('Picture'),
        widget=forms.ClearableFileInput(attrs={'multiple': False})
    )

    birthdate = forms.DateField(
        label=_('Birthdate'),
        widget=forms.TextInput(
            attrs={'type': 'date',
                   'placeholder': _('Birthdate')}))
    """
    country_of_birth = forms.CharField(
        label=_('Country of birth'),
        widget=forms.TextInput(
            attrs={'type': 'text',
                   'placeholder': _('Country of birth')}))
    """
    country_of_birth = CountryField().formfield(label=_('Country of birth'))

    latitude = forms.FloatField(label=_('Latitude'))

    longitude = forms.FloatField(label=_('Longitude'))

    """
    country = forms.CharField(
        label=_('Country'),
        widget=forms.TextInput(
            attrs={'type': 'text',
                   'placeholder': _('Country')}))
    """
    country = CountryField().formfield(label=_('Country'))

    address = forms.CharField(
        label=_('Address'),
        widget=forms.Textarea(
            attrs={'placeholder': _('Address'),
                   'rows': 1}))

    language = forms.CharField(
        label=_('Language'),
        widget=forms.TextInput(
            attrs={'type': 'text',
                   'placeholder': _('Language')}))

    gender = forms.CharField(
        label=_('Gender'),
        widget=forms.TextInput(
            attrs={'type': 'text',
                   'placeholder': _('Gender')}))

    company_name = forms.CharField(
        label=_('Company name'),
        widget=forms.TextInput(
            attrs={'type': 'text',
                   'placeholder': _('Company name')}))

    def set_required_fields(self):
        for field in app_settings.PROFILE_FIELDS:
            if field in self.fields.keys():
                self.fields[field].required = field in app_settings.REQUIRED_PROFILE_FIELDS

    def __init__(self, *args, **kwargs):
        super(ProfileForm, self).__init__(*args, **kwargs)

        self.set_required_fields()

    def clean_picture(self):
        picture_file = self.cleaned_data['picture']
        picture_file_type = type(picture_file)

        if picture_file_type not in (str, None) and picture_file_type in (InMemoryUploadedFile, TemporaryUploadedFile):
            picture_bin = picture_file.file.read()
            picture_image = Image.open(BytesIO(picture_bin))
            picture_format = picture_image.format

            if picture_format not in ('PNG', 'JPEG'):
                raise forms.ValidationError('Image must be either PNG or JPEG')

            picture_image.thumbnail((100, 100), Image.ANTIALIAS)

            image = BytesIO()
            picture_image.save(image, format=picture_format)

            return "data:image/{};base64,".format(picture_format.lower()) + base64.b64encode(image.getvalue()).decode(
                'utf-8')
        else:
            return None


class UserProfileForm(forms.ModelForm, ProfileForm):
    class Meta:
        model = Profile
        fields = app_settings.PROFILE_FIELDS



# Generated by Django 2.2.17 on 2021-03-18 13:56

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_sso_app', '0007_auto_20210222_1526'),
    ]

    operations = [
        migrations.CreateModel(
            name='Grouping',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('path', models.CharField(max_length=255, unique=True)),
                ('depth', models.PositiveIntegerField()),
                ('numchild', models.PositiveIntegerField(default=0)),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('name', models.CharField(max_length=150, verbose_name='name')),
                ('type', models.CharField(blank=True, max_length=255, null=True, verbose_name='type')),
                ('description', models.TextField(blank=True, null=True, verbose_name='description')),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_sso_app.Group')),
            ],
            options={
                'verbose_name': 'Grouping',
            },
        ),
    ]



from django.contrib.auth import authenticate
from django.core.exceptions import ObjectDoesNotExist

from rest_framework import permissions

from . import app_settings


def is_authenticated(user):
    return user is not None and user.is_authenticated


def is_django_staff(user):
    return user and (user.is_staff or user.is_superuser)


def is_staff(user):
    if user is not None:
        if is_django_staff(user):
            return True

        profile = getattr(user, 'sso_app_profile', None)
        if profile is not None:
            return profile.groups.filter(name__in=app_settings.STAFF_USER_GROUPS).count() > 0

    return False


def try_authenticate(username, email, password):
    credentials = username or email or None
    user = None

    if username is None and email is not None:
        credentials = 'email'
    if username is not None and email is None:
        credentials = 'username'

    if credentials is not None:
        user = authenticate(username=credentials, password=password)

    if user is not None:
        return user

    raise ObjectDoesNotExist('Check credentials.')


class OwnerPermission(permissions.IsAuthenticated):
    message = 'You must be the owner.'

    def has_object_permission(self, request, view, obj):
        if (getattr(request, 'user', None) is not None and request.user == obj):
            return True
        return False


class StaffPermission(permissions.IsAuthenticated):
    message = 'You must be a staff member.'

    def has_permission(self, request, view):
        if is_staff(request.user):
            return True
        return False


class NotDjangoStaffUserPermission(permissions.IsAuthenticated):
    message = 'Django staff users are disabled.'

    def has_permission(self, request, view):
        if is_django_staff(request.user):
            return False
        return True


class OwnerOrStaffPermission(permissions.IsAuthenticated):
    message = 'You must be the owner or a staff member.'

    def has_object_permission(self, request, view, obj):
        user = request.user
        if user and (is_staff(user) or (user == getattr(obj, 'user', None))):
            return True
        return False


class PublicObjectOrOwnerOrStaffPermission(permissions.IsAuthenticated):

    def has_object_permission(self, request, view, obj):
        user = request.user
        user_is_staff = is_staff(user)

        if user_is_staff:
            return True

        is_public = getattr(obj, 'is_public', False)
        has_owner = getattr(obj, 'user', False)

        if (is_public and not has_owner) and not user_is_staff:
            return False

        if is_public:
            return True

        obj_user = getattr(obj, 'user', None)

        return obj_user == user



# Generated by Django 2.2.17 on 2021-03-19 13:02

from django.db import migrations, models
import django_countries.fields


class Migration(migrations.Migration):

    dependencies = [
        ('django_sso_app', '0008_grouping'),
    ]

    operations = [
        migrations.AddField(
            model_name='profile',
            name='company_name',
            field=django_countries.fields.CountryField(blank=True, max_length=255, null=True, verbose_name='company name'),
        ),
        migrations.AddField(
            model_name='profile',
            name='country_of_birth',
            field=django_countries.fields.CountryField(blank=True, max_length=46, null=True, verbose_name='country of birth'),
        ),
        migrations.AddField(
            model_name='profile',
            name='expiration_date',
            field=models.DateTimeField(blank=True, null=True, verbose_name='expiration date'),
        ),
        migrations.AddField(
            model_name='profile',
            name='external_id',
            field=models.CharField(blank=True, db_index=True, max_length=36, null=True, verbose_name='external id'),
        ),
        migrations.AddField(
            model_name='profile',
            name='gender',
            field=models.CharField(blank=True, max_length=15, null=True, verbose_name='gender'),
        ),
    ]



from django.urls import reverse

from rest_framework import status

from django_sso_app.core.tests.factories import UserTestCase


class TestTokens(UserTestCase):

    def test_app_shape_uses_settings_secret(self):
        with self.settings(DJANGO_SSO_APP_SHAPE='app', DJANGO_SSO_APP_SERVICE_URL='http://example.com'):
            user = self._get_new_user()
            device = self._get_user_device(user)

            client = self._get_client()
            client.cookies = self._get_valid_jwt_cookie(user.sso_id, user.sso_rev, device.fingerprint)

            response = client.get(
                reverse('django_sso_app_profile:rest-detail', args=(user.sso_app_profile.sso_id,)),
                content_type='application/json'
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.data.get('sso_id'), str(user.sso_app_profile.sso_id))

        with self.settings(DJANGO_SSO_APP_SHAPE='app', DJANGO_SSO_APP_SERVICE_URL='http://example.com'):
            user = self._get_new_user()
            device = self._get_user_device(user)
            invalid_jwt = self._get_jwt(device, 'invalid_secret')

            client = self._get_client()
            client.cookies = self._get_jwt_cookie(device, invalid_jwt)

            response = client.get(
                reverse('django_sso_app_profile:rest-detail', args=(user.sso_app_profile.sso_id,)),
                content_type='application/json'
            )

            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN, 'invalid_jwt is valid')

    def test_app_apigateway_shape_uses_device_secret(self):
        with self.settings(DJANGO_SSO_APP_SHAPE='app_apigateway', DJANGO_SSO_APP_SERVICE_URL='http://example.com'):
            user = self._get_new_user()
            device = self._get_user_device(user)

            client = self._get_client()
            client.cookies = self._get_jwt_cookie(device)

            response = client.get(
                reverse('django_sso_app_profile:rest-detail', args=(user.sso_app_profile.sso_id,)),
                content_type='application/json',
                HTTP_X_CONSUMER_CUSTOM_ID=user.sso_app_profile.sso_id
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.data.get('sso_id'), str(user.sso_app_profile.sso_id))



from django.conf import settings
from meta.views import Meta


def emails_domain_settings(request):
    return {
        "EMAILS_DOMAIN": settings.EMAILS_DOMAIN,
        "EMAILS_SITE_NAME": settings.EMAILS_SITE_NAME,

        "DJANGO_ALLAUTH_ENABLED": getattr(settings, 'DJANGO_ALLAUTH_ENABLED', None)
    }


def google_api_settings(request):
    return {
        "GOOGLE_API_KEY": settings.GOOGLE_API_KEY,
        "GOOGLE_MAPS_API_VERSION": settings.GOOGLE_MAPS_API_VERSION,
        "GOOGLE_ANALYTICS_TRACKING_ID": settings.GOOGLE_ANALYTICS_TRACKING_ID,
    }


def raven_js_dsn_settings(request):
    return {
        "RAVEN_JS_DSN": settings.RAVEN_JS_DSN,
    }


def mapbox_accesstoken(request):
    return {
        "MAPBOX_ACCESSTOKEN": settings.MAPBOX_ACCESSTOKEN,
    }


def disable_js(request):
    is_active = request.GET.get('nojs', None)

    return {
        "DISABLE_JS": is_active is not None,
    }


def get_repository_rev(request):
    return {
        "REPOSITORY_REV": settings.REPOSITORY_REV
    }


def django_meta(request):
    return {
        'meta': Meta(
            description=settings.META_DESCRIPTION,
            extra_props={
                'viewport': 'width=device-width, initial-scale=1.0, minimum-scale=1.0'
            },
            extra_custom_props=[
                ('http-equiv', 'Content-Type', 'text/html; charset=UTF-8'),
            ]
        )
    }


def get_meta_info(request):
    return {
        "META_DESCRIPTION": getattr(settings, 'META_DESCRIPTION', '')
    }



from django.core.management.base import BaseCommand
from django.contrib.auth import get_user_model
from django.core.validators import validate_email
from django.core.exceptions import ValidationError

from django_sso_app.core.apps.users.utils import create_local_user_from_object

User = get_user_model()


class Command(BaseCommand):
    help = 'Loads an external simple user'

    def add_arguments(self, parser):
        parser.add_argument('sso_id', type=str)
        parser.add_argument('username', type=str)
        parser.add_argument('email', type=str)
        parser.add_argument('groups', type=str, nargs='*', default=[])

    def handle(self, *args, **options):
        sso_id = options['sso_id']
        username = options['username']
        email = options['email']
        groups = options['groups']
        groups_count = len(groups)

        try:
            validate_email(email)
        except ValidationError as e:
            self.stdout.write(self.style.ERROR(e.message))
            return

        if User.objects.filter(username=username).count() > 0:
            self.stdout.write(self.style.ERROR('User with username "{}" already present'.format(username)))
            return
        if User.objects.filter(email=email).count() > 0:
            self.stdout.write(self.style.ERROR('User with email "{}" already present'.format(email)))
            return

        self.stdout.write(self.style.NOTICE('Creating user: {} {} {} with {} groups: {}'.format(
            sso_id,
            username,
            email,
            groups_count,
            groups)))

        new_user = create_local_user_from_object({'username': username,
                                                  'email': email,
                                                  'sso_id': sso_id,
                                                  'sso_rev': 0},
                                                 verified=True)

        for group_name in groups:
            new_user.sso_app_profile.add_to_group(group_name, creating=True)

        self.stdout.write(self.style.SUCCESS('New user "{}" created'.format(new_user)))



import logging

# import pyximport
# pyximport.install()

from django.db import transaction

from ...permissions import is_django_staff
from ...functions import lists_differs
from ... import app_settings
from ..services.models import Service
from .models import Profile as Profile

logger = logging.getLogger('django_sso_app.core.apps.profiles')


@transaction.atomic
def create_update_profile_subscriptions(profile, update_object):
    """
    Update local profile subscriptions
    """
    logger.info('Updating profile subscriptions for "{}"'.format(profile))

    remote_subscriptions = update_object.get('subscriptions', None)

    if remote_subscriptions is not None and len(remote_subscriptions) > 0:
        logger.debug('remote_profile_subscriptions "{}"'.format(remote_subscriptions))

        if app_settings.APP_ENABLED:  # disable rev update for app shapes
            setattr(profile.user, '__dssoa__creating', True)

        active_subscriptions = []
        disabled_subscriptions = []
        for s in remote_subscriptions:
            if s['is_unsubscribed']:
                disabled_subscriptions.append(s['service_url'])
            else:
                active_subscriptions.append(s['service_url'])

        active_services = []
        disabled_services = []
        for url in active_subscriptions:
            service, created = Service.objects.get_or_create(service_url=url)
            active_services.append(service)

        for url in disabled_subscriptions:
            service, created = Service.objects.get_or_create(service_url=url)
            disabled_services.append(service)

        """
        active_services = Service.objects.filter(service_url__in=active_subscriptions,
                                                 is_active=True)
        disabled_services = Service.objects.filter(service_url__in=disabled_subscriptions,
                                                   is_active=True)
        """

        active_subscriptions = []
        disabled_subscriptions = []
        for s in active_services:
            subscription, updated = s.subscribe(profile)
            if updated:
                logger.info('Subscription updated "{}"'.format(subscription))
                active_subscriptions.append(subscription)

        for s in disabled_services:
            subscription, updated = s.unsubscribe(profile)
            if updated:
                logger.info('Subscription updated "{}"'.format(subscription))
                disabled_subscriptions.append(subscription)

        if app_settings.APP_ENABLED:  # restore rev update for app shapes
            setattr(profile.user, '__dssoa__creating', False)

        # return
        return active_subscriptions, disabled_subscriptions

    return None, None


@transaction.atomic
def update_profile_groups(profile, update_object):
    logger.debug('update_profile_groups')

    actual_groups_list = update_object.get('groups', None)

    if actual_groups_list is not None:
        if app_settings.APP_ENABLED:  # disable rev update for app shapes
            setattr(profile.user, '__dssoa__creating', True)

        active_groups = []
        disabled_groups = []

        previous_groups_list = list(profile.groups.order_by('name').values_list('name', flat=True))

        if lists_differs(previous_groups_list, actual_groups_list):
            logger.info('Must update profile groups for "{}"'.format(profile))
            logger.info('From {} to {}'.format(previous_groups_list, actual_groups_list))

            # try to remove profile from removed groups
            for group_name in previous_groups_list:
                if group_name not in actual_groups_list:
                    profile.remove_from_group(group_name)
                    disabled_groups.append(group_name)

            # try to add profile to new groups
            for group_name in actual_groups_list:
                if group_name not in previous_groups_list:
                    profile.add_to_group(group_name)
                    active_groups.append(group_name)

        if app_settings.APP_ENABLED:  # disable rev update for app shapes
            setattr(profile.user, '__dssoa__creating', True)

        return active_groups, disabled_groups

    return None, None


@transaction.atomic
def update_profile(profile, update_object, commit=True):
    logger.info('Updating profile fields for "{}"'.format(profile))

    for f in app_settings.PROFILE_FIELDS + ('sso_rev', 'meta'):
        if hasattr(profile, f):
            new_val = update_object.get(f, None)
            if new_val is not None:
                setattr(profile, f, new_val)
                logger.debug('profile field updated "{}":"{}"'.format(f, new_val if f != 'password' else '*'))

    if commit:
        profile.save()

    if profile.pk:  # if profile model is in DB
        logger.debug('Try update profile groups and subscriptions')
        # update groups
        _active_groups, _disabled_groups = update_profile_groups(profile, update_object)
        # update subscriptions
        _active_subscriptions, _disabled_subscriptions = create_update_profile_subscriptions(profile, update_object)

    return profile


def get_or_create_user_profile(user, initials=None, commit=True):
    logger.debug('get or create user profile')

    try:
        profile = user.sso_app_profile

    except Profile.DoesNotExist:
        logger.info('User "{}" has no profile, creating one'.format(user))

        setattr(user, '__dssoa__creating', True)

        profile = Profile(user=user, django_user_email=user.email, django_user_username=user.username)

        sso_id = getattr(user, '__dssoa__profile__sso_id', None)
        if sso_id is not None:
            sso_rev = getattr(user, '__dssoa__profile__sso_rev', 0)
            profile.sso_id = sso_id
            profile.sso_rev = sso_rev
        else:
            # django staff users sso_id equals username
            if is_django_staff(user):
                profile.sso_id = user.username
                profile.sso_rev = 0

        profile_initials = initials or getattr(user, '__dssoa__profile__object', None)

        if profile_initials is not None:
            logger.debug('Got profile initials "{}"'.format(profile_initials))
            profile = update_profile(profile, profile_initials, commit=commit)
        else:
            logger.debug('No profile initials')
            if commit:
                profile.save()

        setattr(user, '__dssoa__creating', False)

    # force user sso_app_profile object field
    #setattr(user, 'sso_app_profile', profile)

    return profile



from django.contrib import admin

from .models import Device


class DeviceAdmin(admin.ModelAdmin):
    search_fields = ('id', 'fingerprint', 'profile__sso_id', 'profile__user__username', 'profile__user__email')
    list_display = ('user', 'fingerprint', 'created_at')


admin.site.register(Device, DeviceAdmin)



import json

from datetime import timezone

from celery import shared_task
from celery.utils.log import get_task_logger

from django.contrib.auth import get_user_model
from django.db import transaction
from django.utils import translation
from django.conf import settings

from allauth.account.models import EmailAddress

from ..profiles.models import Profile
from ..groups.models import Group
from ..users.utils import create_local_user_from_object
from .models import RequestForCredentialsEventListener
from .utils import fetch_event_type_events, sort_event_type_events

User = get_user_model()
MODULE_NAME = 'events'


# https://stackoverflow.com/questions/12414821/checking-a-nested-dictionary-using-a-dot-notation-string-a-b-c-d-e-automatica
def _put(d, keys, item):
    if "." in keys:
        key, rest = keys.split(".", 1)
        if key not in d:
            d[key] = {}
        _put(d[key], rest, item)
    else:
        d[keys] = item


def _get(d, keys):
    if "." in keys:
        key, rest = keys.split(".", 1)
        return _get(d[key], rest)
    else:
        return d.get(keys, None)


def _create_users_by_events(logger):
    """
    Check user related events
    :param logger:
    :return:
    """
    users_created = 0

    instantiated_event_listeners = RequestForCredentialsEventListener.objects.filter(is_active=True)

    for event_listener in instantiated_event_listeners:
        last_profile = Profile.objects.filter(created_by_event=True).order_by('-created_at').first()
        from_date = last_profile.created_at if last_profile else event_listener.created_at
        from_timestamp = from_date.timestamp()
        event_type = event_listener.event_type
        confirm_email = event_listener.confirm_email
        default_password = event_listener.default_password

        logger.info('Getting "{}" events from date "{}"'.format(event_type, from_date))
        rfc_events, _count = fetch_event_type_events(event_type, from_timestamp)

        for event in sort_event_type_events(rfc_events, event_type):
            try:
                with transaction.atomic():
                    logger.debug('Last "{}" event: "{}"'.format(event_type, event))

                    # set email
                    user_email = event.get(event_listener.email_field, None)
                    if user_email is None:
                        raise Exception('Event object has no "{}" field'.format(event_listener.email_field))

                    # set username
                    user_username = event.get(event_listener.email_field, user_email)

                    profile_object = {}
                    for field in RequestForCredentialsEventListener.profile_fields:
                        field_key = getattr(event_listener, '{}_field'.format(field), None)
                        if field_key is not None:
                            # val = event.get(field_key, None)
                            val = _get(event, field_key)
                            logger.debug('Event profile field {}:{}'.format(field_key, val))
                            if val is not None:
                                profile_object[field] = val

                    # create user
                    new_user = create_local_user_from_object({
                        'email': user_email,
                        'username': user_username,
                        'profile': profile_object
                    }, verified=confirm_email)

                    # set created by event
                    new_user.sso_app_profile.created_by_event = True
                    new_user.sso_app_profile.meta = json.dumps({
                        'event': event
                    })
                    new_user.sso_app_profile.save()

                    # set profile groups
                    profile_group_field_key = getattr(event_listener, 'group_field', None)
                    if profile_group_field_key is not None:
                        profile_group_name = event.get(profile_group_field_key, None)
                        if profile_group_name is not None:
                            new_user.sso_app_profile.add_to_group(profile_group_name, creating=True)

                    # set user password
                    if default_password is not None:
                        logger.info('Setting default password')
                        new_user.set_password(default_password)
                        new_user.save()

                    # send confirmation email
                    if not confirm_email:
                        logger.info('Sending confirmation email to"{}"'.format(user_email))
                        language_code = new_user.sso_app_profile.language or settings.LANGUAGE_CODE
                        translation.activate(language_code)
                        EmailAddress.objects.get(email=user_email).send_confirmation()

                    logger.info('New user: "{}"'.format(new_user))
                    users_created += 1

            except Exception as e:
                logger.exception('Error: {}. Can not create user by event: {}'.format(e, event))

    return users_created


@shared_task(bind=True)
def create_users_by_events(self, **kwargs):
    _logger = kwargs.get('logger', get_task_logger('{}.{}'.format(MODULE_NAME, 'create_users_by_events')))

    return _create_users_by_events(_logger)



import logging

from django.contrib.auth import get_user_model
from django import forms

from allauth.account.adapter import get_adapter
from rest_framework import serializers

from ...utils import set_session_key
from ... import app_settings
from ...permissions import is_staff
from ...serializers import AbsoluteUrlSerializer
from ..emails.models import EmailAddress
from ..profiles.models import Profile
from ..profiles.utils import update_profile
from ..profiles.serializers import ProfileSerializer
from ..services.models import Service
from .utils import create_local_user_from_object

User = get_user_model()

logger = logging.getLogger('django_sso_app.core.apps.users')


class CheckUserExistenceSerializer(serializers.ModelSerializer):
    sso_id = serializers.SerializerMethodField(required=False)

    class Meta:
        model = User
        fields = ('sso_id',)
        read_only_fields = ('sso_id', )

    def get_sso_id(self, instance):
        return instance.sso_id


class UserRevisionSerializer(serializers.ModelSerializer):
    sso_id = serializers.SerializerMethodField(required=False)
    sso_rev = serializers.SerializerMethodField(required=False)

    class Meta:
        model = User
        read_only_fields = ('sso_id', 'sso_rev')
        fields = read_only_fields

    def get_sso_id(self, instance):
        return instance.sso_id

    def get_sso_rev(self, instance):
        return instance.sso_rev


class UserProfileSerializer(AbsoluteUrlSerializer):
    sso_id = serializers.CharField(required=False)
    sso_rev = serializers.IntegerField(required=False)

    subscriptions = serializers.SerializerMethodField(required=False)

    ssn = serializers.CharField(required='ssn' in app_settings.REQUIRED_PROFILE_FIELDS)
    phone = serializers.CharField(required='phone' in app_settings.REQUIRED_PROFILE_FIELDS)

    first_name = serializers.CharField(required='first_name' in app_settings.REQUIRED_PROFILE_FIELDS)
    last_name = serializers.CharField(required='last_name' in app_settings.REQUIRED_PROFILE_FIELDS,
                                      allow_null=True,
                                      allow_blank=True)

    description = serializers.CharField(required='description' in app_settings.REQUIRED_PROFILE_FIELDS,
                                        allow_null=True, allow_blank=True)
    picture = serializers.CharField(required='picture' in app_settings.REQUIRED_PROFILE_FIELDS, allow_null=True)
    birthdate = serializers.DateField(required='birthdate' in app_settings.REQUIRED_PROFILE_FIELDS,
                                      allow_null=True)
    country_of_birth = serializers.CharField(required='country_of_birth' in app_settings.REQUIRED_PROFILE_FIELDS)

    latitude = serializers.FloatField(required='latitude' in app_settings.REQUIRED_PROFILE_FIELDS)
    longitude = serializers.FloatField(required='longitude' in app_settings.REQUIRED_PROFILE_FIELDS)

    country = serializers.CharField(required='country' in app_settings.REQUIRED_PROFILE_FIELDS)
    address = serializers.CharField(required='address' in app_settings.REQUIRED_PROFILE_FIELDS, allow_null=True)

    language = serializers.CharField(required='language' in app_settings.REQUIRED_PROFILE_FIELDS,
                                     allow_null=True)

    alias = serializers.CharField(required='alias' in app_settings.REQUIRED_PROFILE_FIELDS,
                                  allow_null=True)

    gender = serializers.CharField(required='gender' in app_settings.REQUIRED_PROFILE_FIELDS,
                                   allow_null=True)

    company_name = serializers.CharField(required='company_name' in app_settings.REQUIRED_PROFILE_FIELDS)

    class Meta:
        model = Profile
        read_only_fields = (
            'url',
            'created_at',
            'sso_id', 'sso_rev',
            'subscriptions',
            'is_unsubscribed',
            'groups')
        fields = read_only_fields + app_settings.PROFILE_FIELDS

    def get_subscriptions(self, instance):
        serialized = []
        if app_settings.BACKEND_ENABLED:
            for el in instance.subscriptions.all():
                serialized.append({
                    'service_url': el.service.service_url,
                    'is_unsubscribed': getattr(el, 'is_unsubscribed', False)
                })
        return serialized


class UserSerializer(AbsoluteUrlSerializer):
    sso_id = serializers.SerializerMethodField(required=False)
    sso_rev = serializers.SerializerMethodField(required=False)
    email_verified = serializers.SerializerMethodField(required=False)
    groups = serializers.SerializerMethodField(required=False)
    profile = ProfileSerializer(required=False, source='sso_app_profile')

    class Meta:
        model = User
        read_only_fields = (
            'url',
            'sso_id', 'sso_rev',
            'date_joined', 'last_login',
            'is_active', 'email_verified',
            'groups')
        fields = read_only_fields + app_settings.USER_FIELDS + ('password', 'profile')

    def get_sso_id(self, instance):
        return instance.sso_id

    def get_sso_rev(self, instance):
        return instance.sso_rev

    def get_email_verified(self, user):
        if app_settings.BACKEND_ENABLED:
            try:
                verified_email_addresses = user.emailaddress_set.filter(email=user.email, verified=True)
                return verified_email_addresses.count() > 0

            except:
                logger.info('user "{}" has no verified emails'.format(user))
                return False

        return True

    def get_groups(self, instance):
        groups = list(instance.groups.values_list('name', flat=True))

        return groups

    def to_representation(self, obj):
        # get the original representation
        ret = super(UserSerializer, self).to_representation(obj)
        request = self.context.get('request', None)

        if request is not None:
            # remove field if password if not asked
            requesting_user = request.user
            requesting_user_is_staff = is_staff(requesting_user)
            is_same_user = requesting_user.sso_id == ret['sso_id']

            with_password = request.query_params.get('with_password',
                                                     not app_settings.BACKEND_HIDE_PASSWORD_FROM_USER_SERIALIZER)

            show_user_protected_informations = False
            hide_password = True

            if requesting_user_is_staff or is_same_user:
                show_user_protected_informations = True

            if show_user_protected_informations:
                logger.info('show protected informations')
                if requesting_user_is_staff and with_password:
                    hide_password = False
            else:
                ret.pop('email', None)

                ret['profile'].pop('email', None)
                ret['profile'].pop('first_name', None)
                ret['profile'].pop('last_name', None)
                ret['profile']['_partial'] = True

            if hide_password:
                ret.pop('password', None)
                ret['_partial'] = True

        # return the modified representation
        return ret

    def update(self, instance, validated_data):
        user = instance
        request = self.context.get("request")
        requesting_user = request.user
        adapter = get_adapter(request=request)

        requesting_user_is_staff = is_staff(requesting_user)

        skip_confirmation = request.GET.get('skip_confirmation', None)
        must_confirm_email = True
        if requesting_user_is_staff and skip_confirmation is not None:
            must_confirm_email = False

        password_is_hashed = request.query_params.get('password_is_hashed', None)

        new_email = validated_data.get('email', None)
        new_password = validated_data.get('password', None)
        new_username = validated_data.get('username', None)
        updated_profile = request.data.get('profile', None)

        logger.info('User "{}" wants to update user "{}"'.format(requesting_user, user))

        if (new_email is not None) and (new_email != user.email):

            logger.info('User "{}" is changing email for user "{}"'.format(requesting_user,
                                                                           user))

            adapter.validate_unique_email(new_email)

            created_email = EmailAddress.objects.add_email(request, user,
                                                           new_email,
                                                           confirm=must_confirm_email)

            logger.info('A new user EmailAddress for user "{}" has been created ({})'.format(user,
                                                                                             created_email.id))

            if must_confirm_email:
                logger.info('"{}" must confirm email "{}"'.format(user, new_email))

                user.sso_app_profile.update_rev(True)

            else:
                logger.info('confirming email "{}"'.format(new_email))
                created_email.verified = True
                created_email.set_as_primary(conditional=False)

                adapter.update_user_email(user, new_email)

        if new_password is not None:
            logger.info('user "{}" is changing password for user "{}"'.format(requesting_user,
                                                                              user))

            if password_is_hashed:
                logger.info('password is hashed')
                user.password = new_password
            else:
                logger.info('password is plain')
                user.set_password(new_password)

            set_session_key(request, '__dssoa__user_password_updated', True)

            user.save()

        if new_username is not None and not user.sso_app_profile.completed_at:
            logger.info('user "{}" is changing username for user "{}"'.format(requesting_user,
                                                                              user))

            try:
                adapter.clean_username(new_username)

            except forms.ValidationError:
                logger.warning('Can\'t set username as "{}" for user "{}"'.format(new_username, user))
                raise

            else:
                adapter.update_user_username(user, new_username)

        if updated_profile is not None:
            logger.info('user "{}" is updating profile for user "{}"'.format(requesting_user,
                                                                             user))

            # setting new_user profile serializer field
            setattr(user, 'sso_app_profile', update_profile(user.sso_app_profile, updated_profile))

        return user

    def get_cleaned_data(self):
        return self.validated_data

    @property
    def cleaned_data(self):
        return self.validated_data


class NewUserSerializer(UserSerializer):
    profile = UserProfileSerializer(required=False)

    def create(self, validated_data):
        request = self.context.get('request')
        requesting_user = request.user

        requesting_user_is_staff = is_staff(requesting_user)
        skip_confirmation = request.GET.get('skip_confirmation', None)
        must_confirm_email = True
        password_is_hashed = request.query_params.get('password_is_hashed', None)

        if password_is_hashed:
            new_password = validated_data.pop('password', None)
        else:
            new_password = validated_data.get('password', None)
            validated_data.update(password1=new_password)
            validated_data.update(password2=new_password)

        assert (new_password is not None)

        # start creating new user
        new_user = None
        try:
            logger.info('User "{}" wants to create new user "{}"'.format(requesting_user, validated_data))

            if requesting_user_is_staff and skip_confirmation is not None:
                must_confirm_email = False

            new_user = create_local_user_from_object(validated_data, verified=(not must_confirm_email))
            new_user_profile_validated_object = validated_data.get('profile', {})

            # setting new_user profile serializer field
            setattr(new_user, 'profile', update_profile(new_user.sso_app_profile, new_user_profile_validated_object))

            if password_is_hashed:
                logger.info('New user password is hashed')
                new_user.password = new_password
                new_user.save()

            try:
                new_user_profile_request_object = request.data.get('profile', {})
                profile_object_subscriptions = new_user_profile_request_object.get('subscriptions', [])

                logger.info('Subscripting new user to services "{}"'.format(profile_object_subscriptions))

                if len(profile_object_subscriptions) > 0:
                    for new_subscription in profile_object_subscriptions:
                        new_subscription_service_name = new_subscription.get('service').get('name')
                        service = Service.objects.get(name=new_subscription_service_name)
                        service.subscribe(new_user.get_sso_app_profile())

            except Exception:
                logger.exception('Error subscribing service on user creation')
                raise serializers.ValidationError('Error subscribing service')

            if must_confirm_email:
                logger.info('Sending confirmation email to new user "{}"'.format(new_user))
                new_user_email_address = new_user.emailaddress_set.last()
                new_user_email_address.send_confirmation(request=request, signup=True)

            return new_user

        # delete creating user model if any exception occurred
        except Exception as e:
            logger.exception('Error creating new user "{}"'.format(e))

            if new_user is not None:
                new_user.delete()

            raise


class SuccessfullLoginResponseSerializer(serializers.Serializer):
    """
    Login success response serializer
    """
    token = serializers.CharField()
    user = UserSerializer()
    redirect_url = serializers.CharField(required=False)


class SuccessfullLogoutResponseSerializer(serializers.Serializer):
    """
    Logout success response serializer
    """
    redirect_url = serializers.CharField(required=False)



from importlib import import_module

from django.conf.urls import url
from django.contrib.auth.decorators import login_required
from django.urls import include
from django.views.generic.base import RedirectView

from allauth import app_settings
from allauth.socialaccount import providers
# from allauth.socialaccount.providers.google.urls import urlpatterns as allauth_aocialaccount_google_urlpatterns

from .views import SignupView, LoginView, LogoutView, PasswordChangeView, PasswordSetView, \
                   AccountInactiveView, EmailView, EmailVerificationSentView, ConfirmEmailView, PasswordResetView, \
                   PasswordResetDoneView, PasswordResetFromKeyView, PasswordResetFromKeyDoneView, \
                   SocialSignupView, SocialLoginCancelledView, SocialLoginErrorView, SocialConnectionsView


allauth_i18n_urlpatterns = [
    url(r"^signup/$", SignupView.as_view(), name="account_signup"),
    url(r"^login/$", LoginView.as_view(), name="account_login"),
    url(r"^logout/$", LogoutView.as_view(), name="account_logout"),
]

allauth_urlpatterns = [
    # Password
    url(r"^password/change/$", login_required(PasswordChangeView.as_view()),
        name="account_change_password"),
    url(r"^password/set/$", login_required(PasswordSetView.as_view()), name="account_set_password"),

    # password reset
    url(r"^password/reset/$", PasswordResetView.as_view(),
        name="account_reset_password"),
    url(r"^password/reset/done/$", PasswordResetDoneView.as_view(),
        name="account_reset_password_done"),

    # SPA frontend required view
    url(r"^password/reset/key/$", RedirectView.as_view(url='/password/reset/', permanent=False),
        name="account_reset_password_key_redirect"),

    url(r"^password/reset/key/(?P<uidb36>[0-9A-Za-z]+)/(?P<key>.+)/$",
        PasswordResetFromKeyView.as_view(),
        name="account_reset_password_from_key"),
    url(r"^password/reset/key/done/$", PasswordResetFromKeyDoneView.as_view(),
        name="account_reset_password_from_key_done"),

    # E-mail
    url(r"^email/$", login_required(EmailView.as_view()), name="account_email"),

    url(r"^confirm-email/$", EmailVerificationSentView.as_view(),
        name="account_email_verification_sent"),

    url(r"^inactive/$", AccountInactiveView.as_view(), name="account_inactive"),

    url(r"^confirm-email/(?P<key>[-:\w]+)/$", ConfirmEmailView.as_view(),
        name="account_confirm_email"),

    # completely overridden
    # url(r'', include('allauth.urls')),
]

if app_settings.SOCIALACCOUNT_ENABLED:
    allauth_socialaccount_urlpatterns = []

    allauth_socialaccount_urlpatterns += [
        url(r'^login/cancelled/$', SocialLoginCancelledView.as_view(),
            name='socialaccount_login_cancelled'),
        url(r'^login/error/$', SocialLoginErrorView.as_view(),
            name='socialaccount_login_error'),
        url(r'^signup/$', SocialSignupView.as_view(), name='socialaccount_signup'),
        url(r'^connections/$', SocialConnectionsView.as_view(), name='socialaccount_connections'),

    ]

    # taken from allauth.urls
    # Provider urlpatterns, as separate attribute (for reusability).
    provider_urlpatterns = []
    for provider in providers.registry.get_list():
        try:
            prov_mod = import_module(provider.get_package() + '.urls')
        except ImportError:
            continue
        prov_urlpatterns = getattr(prov_mod, 'urlpatterns', None)
        if prov_urlpatterns:
            allauth_socialaccount_urlpatterns += prov_urlpatterns

    allauth_urlpatterns += [
        url(r'^social/', include(allauth_socialaccount_urlpatterns)),
    ]

    # completely overridden
    # url(r'^social/', include('allauth.socialaccount.urls'))



import logging

from django.contrib.auth import get_user_model

from rest_framework.authentication import SessionAuthentication

User = get_user_model()

logger = logging.getLogger('django_sso_app.core.api')


class DjangoSsoApiAuthentication(SessionAuthentication):
    """
    Returns authenticated user if set by authentication middleware
    """

    def enforce_csrf(self, request):
        """ SKIP Enforce CSRF validation for session based authentication. """
        pass



"""from django.contrib import admin

from .models import Status


@admin.register(Status)
class StatusAdmin(admin.ModelAdmin):
    list_display = ('shape', 'created_at')


admin.site.register(Status, StatusAdmin)
"""


from django.contrib import admin

from .models import RequestForCredentialsEventListener


class RequestForCredentialsEventListenerAdmin(admin.ModelAdmin):
    list_display = ('event_type', 'created_at', 'is_active')
    search_fields = ('event_type',)


admin.site.register(RequestForCredentialsEventListener, RequestForCredentialsEventListenerAdmin)



# Generated by Django 2.2.9 on 2020-03-07 10:35

from django.db import migrations, models
import django_countries.fields


class Migration(migrations.Migration):

    dependencies = [
        ('django_sso_app', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='profile',
            name='groups',
            field=models.ManyToManyField(blank=True, related_name='sso_app_profiles', related_query_name='sso_app_profile', to='django_sso_app.Group', verbose_name='groups'),
        ),
        migrations.AlterField(
            model_name='profile',
            name='country',
            field=django_countries.fields.CountryField(blank=True, max_length=46, null=True, verbose_name='country'),
        ),
        migrations.AlterField(
            model_name='service',
            name='groups',
            field=models.ManyToManyField(blank=True, related_name='sso_app_services', related_query_name='sso_app_service', to='django_sso_app.Group', verbose_name='groups'),
        ),
    ]



import logging
import os

# from django.conf import settings
from django.core.exceptions import SuspiciousOperation
from django.shortcuts import redirect
from django.urls import reverse
from django.http import HttpResponse
from django.template import loader
from django.contrib.auth import logout as django_logout

from rest_framework import status

from allauth.account import views as allauth_account_views
from allauth.account.adapter import get_adapter
from allauth.socialaccount import views as allauth_socialaccount_views

from .apps.devices.utils import renew_response_jwt
from .permissions import is_django_staff, is_authenticated, is_staff
from .tokens.utils import get_request_jwt
from .utils import set_cookie, invalidate_cookie, set_session_key, get_session_key, get_random_fingerprint
from .mixins import WebpackBuiltTemplateViewMixin
from . import app_settings

logger = logging.getLogger('django_sso_app.core')

CURRENT_DIR = os.getcwd()


class DjangoSsoAppBaseViewMixin:
    def update_response_jwt(self, user, response):
        """
        Updated response JWT (if in request)
        """
        if self.request is None:
            raise KeyError('no request object')

        requesting_user = self.request.user

        if not is_django_staff(requesting_user):
            logger.info('Updating response jwt for "{}"'.format(user))
            received_jwt = get_request_jwt(self.request)

            if received_jwt is not None:
                is_same_user = requesting_user == user

                if is_same_user:
                    logger.info('From browser as same user "{}"'.format(user))
                    renew_response_jwt(received_jwt, user, self.request, response)

                elif not is_same_user and is_staff(requesting_user):
                    # returns user jwt to staff user  #Q!
                    logger.info('From browser as staff user "{}"'.format(requesting_user))
                    renew_response_jwt(received_jwt, user, self.request, response)

                else:
                    _msg = 'User "{}" can not update response JWT for "{}"'.format(requesting_user, user)
                    logger.warning(_msg)

                    raise SuspiciousOperation(_msg)


class DjangoSsoAppViewMixin(DjangoSsoAppBaseViewMixin):
    def __init__(self, *args, **kwargs):
        super(DjangoSsoAppViewMixin, self).__init__(*args, **kwargs)

        if not hasattr(self, 'object'):
            setattr(self, 'object', None)

    @staticmethod
    def get_request_fingerprint(request):
        return request.POST.get('fingerprint', get_random_fingerprint(request))

    @staticmethod
    def get_response_with_cookie(request, response):
        if is_authenticated(request.user):
            if not is_django_staff(request.user):
                token = get_session_key(request, '__dssoa__jwt_token')
                set_cookie(response, app_settings.JWT_COOKIE_NAME, token)

        return response

    @staticmethod
    def get_response_with_invalidated_cookie(response):
        invalidate_cookie(response, app_settings.JWT_COOKIE_NAME)

        return response


# allauth views

class LoginView(allauth_account_views.LoginView, DjangoSsoAppViewMixin, WebpackBuiltTemplateViewMixin):

    def get(self, request, *args, **kwargs):
        logger.info('Getting login')

        try:
            return super(LoginView, self).get(request, *args, **kwargs)

        except:
            logger.exception('Error getting login')
            raise

    def post(self, request, *args, **kwargs):
        logger.info('Logging in')

        try:
            set_session_key(request, '__dssoa__device__fingerprint', self.get_request_fingerprint(request))
            response = super(LoginView, self).post(request, *args, **kwargs)
            location = response.get('location', None)
            if location == reverse('account_email_verification_sent'):
                # disable login on signup
                logger.info('User "must confirm email"'.format(request.user))

                adapter = get_adapter(self.request)
                adapter.logout(request)

        except:
            logger.exception('Error logging in')
            raise

        else:
            if is_authenticated(request.user):
                logger.info('login success')
                response = self.get_response_with_cookie(request, response)

            return response


class LogoutView(allauth_account_views.LogoutView, DjangoSsoAppViewMixin, WebpackBuiltTemplateViewMixin):
    def get(self, request, *args, **kwargs):
        logger.info('Getting logout')

        if request.user and request.user.is_anonymous:
            return redirect(reverse('account_login'))

        try:
            return super(LogoutView, self).get(request, *args, **kwargs)

        except:
            logger.exception('Error getting logout')
            raise

    def post(self, request, *args, **kwargs):
        logger.info('Logging out')

        if request.user and request.user.is_anonymous:
            return redirect(reverse('account_login'))

        try:
            response = super(LogoutView, self).post(request, *args, **kwargs)

        except:
            logger.exception('Error logging out')
            raise

        else:
            return self.get_response_with_invalidated_cookie(response)


class SignupView(allauth_account_views.SignupView, DjangoSsoAppViewMixin, WebpackBuiltTemplateViewMixin):
    def get(self, request,  *args, **kwargs):
        logger.info('Getting signup')

        if is_authenticated(request.user):
            return redirect(reverse('account_login'))

        try:
            return super(SignupView, self).get(request, *args, **kwargs)

        except:
            logger.exception('Error getting signup')
            raise

    def post(self, request, *args, **kwargs):
        logger.info('Signing up')

        if is_authenticated(request.user):
            return redirect(reverse('account_login'))

        try:
            response = super(SignupView, self).post(request, *args, **kwargs)

        except:
            logger.exception('Error signin up')
            raise

        else:
            # log signup failed
            if response.status_code not in (200, 302):
                if hasattr(response, 'render'):
                    response.render()
                logger.warning('User failed to register: "{}"'.format(response.content))

            if is_authenticated(request.user):
                logger.debug('signup login disabled')
                django_logout(request)

        return response


class ConfirmEmailView(allauth_account_views.ConfirmEmailView, DjangoSsoAppViewMixin):
    def get(self, request, *args, **kwargs):
        logger.info('Getting confirm email')

        try:
            response = super(ConfirmEmailView, self).get(request, *args, **kwargs)

        except:
            logger.exception('Error get confirm email')
            raise

        else:
            if is_authenticated(request.user):
                response = self.get_response_with_cookie(request, response)

            return response

    def post(self, request, *args, **kwargs):
        logger.info('Confirming email')

        try:
            response = super(ConfirmEmailView, self).post(request, *args, **kwargs)

        except:
            logger.exception('Error confirming email')
            raise

        else:
            if is_authenticated(request.user):
                response = self.get_response_with_cookie(request, response)

            return response


class EmailView(allauth_account_views.EmailView):
    def get(self, *args, **kwargs):
        logger.info('Getting email')

        try:
            return super(EmailView, self).get(*args, **kwargs)

        except:
            logger.exception('Error getting email')
            raise

    def post(self, request, *args, **kwargs):
        logger.info('Posting email')

        try:
            response = super(EmailView, self).post(request, *args, **kwargs)

        except:
            logger.exception('Error email')
            raise

        return response


class PasswordChangeView(allauth_account_views.PasswordChangeView):
    def get(self, *args, **kwargs):
        logger.info('Getting password change')

        try:
            return super(PasswordChangeView, self).get(*args, **kwargs)

        except:
            logger.exception('Error getting password change')
            raise

    def post(self, *args, **kwargs):
        logger.info('Changing password')

        try:
            return super(PasswordChangeView, self).post(*args, **kwargs)

        except:
            logger.exception('Error changing password')
            raise


class PasswordSetView(allauth_account_views.PasswordSetView, DjangoSsoAppViewMixin):
    def get(self, *args, **kwargs):
        logger.info('Getting password set')

        try:
            return super(PasswordSetView, self).get(*args, **kwargs)

        except:
            logger.exception('Error getting password set')
            raise

    def post(self, request, *args, **kwargs):
        logger.info('Setting password')

        try:
            response = super(PasswordSetView, self).post(request, *args, **kwargs)

        except:
            logger.exception('Error setting password')
            raise

        else:
            self.update_response_jwt(request.user, response)

            return response


class PasswordResetView(allauth_account_views.PasswordResetView, DjangoSsoAppViewMixin, WebpackBuiltTemplateViewMixin):
    def get(self, *args, **kwargs):
        logger.info('Getting ask password reset')

        try:
            return super(PasswordResetView, self).get(*args, **kwargs)

        except:
            logger.exception('Error getting ask password reset')
            raise

    def post(self, *args, **kwargs):
        logger.info('Asking for password reset')

        try:
            return super(PasswordResetView, self).post(*args, **kwargs)

        except:
            logger.exception('Error while asking password reset')
            raise


class PasswordResetDoneView(allauth_account_views.PasswordResetDoneView):
    def get(self, request, *args, **kwargs):
        logger.info('Getting password reset done')

        try:
           return super(PasswordResetDoneView, self).get(request, *args, **kwargs)

        except:
            logger.exception('Error getting password reset done')
            raise


class PasswordResetFromKeyView(allauth_account_views.PasswordResetFromKeyView, DjangoSsoAppViewMixin, WebpackBuiltTemplateViewMixin):

    def dispatch(self, request, uidb36, key, **kwargs):
        _response = super(PasswordResetFromKeyView, self).dispatch(request, uidb36, key, **kwargs)
        _status_code = _response.status_code

        if app_settings.BACKEND_FRONTEND_APP_RESTONLY and _status_code == status.HTTP_302_FOUND:
            template_name = self.get_template_names()[0]
            logger.debug('Frontend is rest only, returning template "{}"'.format(template_name))
            template = loader.get_template(template_name)
            context = {}

            return HttpResponse(template.render(context, request))

        return _response

    def get(self, request, uidb36, key, **kwargs):
        logger.info('Getting password reset from key')

        try:
            return super(PasswordResetFromKeyView, self).get(request, uidb36, key, **kwargs)

        except:
            logger.exception('Error while getting password reset from key')
            raise

    def post(self, request, uidb36, key, **kwargs):
        logger.info('Resetting password form key')

        try:
            return super(PasswordResetFromKeyView, self).post(request, uidb36, key, **kwargs)

        except:
            logger.exception('Error while resetting password form key')
            raise


class PasswordResetFromKeyDoneView(allauth_account_views.PasswordResetFromKeyDoneView):
    def get(self, *args, **kwargs):
        logger.info('Getting password reset from key done')
        try:
            return super(PasswordResetFromKeyDoneView, self).get(*args,
                                                                 **kwargs)
        except:
            logger.exception('Error while getting password reset from key done')
            raise


class AccountInactiveView(allauth_account_views.AccountInactiveView):
    def get(self, *args, **kwargs):
        logger.info('Getting account inactive')

        try:
            return super(AccountInactiveView, self).get(*args, **kwargs)

        except:
            logger.exception('Error while getting account inactive')
            raise


class EmailVerificationSentView(allauth_account_views.EmailVerificationSentView, DjangoSsoAppViewMixin, WebpackBuiltTemplateViewMixin):
    def get(self, *args, **kwargs):
        logger.info('Getting email verification sent')

        try:
            return super(EmailVerificationSentView, self).get(*args, **kwargs)

        except:
            logger.exception('Error getting verification sent')
            raise


# allauth socialaccount views
class SocialSignupView(allauth_socialaccount_views.SignupView):
    pass


class SocialLoginCancelledView(allauth_socialaccount_views.LoginCancelledView):
    pass


class SocialLoginErrorView(allauth_socialaccount_views.LoginErrorView):
    pass


class SocialConnectionsView(allauth_socialaccount_views.ConnectionsView):
    pass



import pytest
from django.urls import resolve, reverse

from django_sso_app.backend.users.models import User

pytestmark = pytest.mark.django_db


def test_detail(user: User):
    assert (
        reverse("users:detail", kwargs={"username": user.username})
        == f"/users/{user.username}/"
    )
    assert resolve(f"/users/{user.username}/").view_name == "users:detail"


def test_update():
    assert reverse("users:update") == "/users/~update/"
    assert resolve("/users/~update/").view_name == "users:update"


def test_redirect():
    assert reverse("users:redirect") == "/users/~redirect/"
    assert resolve("/users/~redirect/").view_name == "users:redirect"



import logging

import requests

from ... import app_settings

logger = logging.getLogger('django_sso_app.core.apps.services')


def subscribe_to_backend_service(sso_id, encoded_jwt=None):
    """
    Creates a new profile subscription

    :param sso_id:
    :return:
    """
    logger.info("Subscribing profile with SSO ID {sso_id} to service"
                " {service_name} ..."
                .format(sso_id=sso_id,
                        service_name=app_settings.SERVICE_URL))
    if encoded_jwt is None:
        headers = {
            "Content-Type": "application/json",
            "Authorization": "Token {}".format(app_settings.BACKEND_STAFF_TOKEN)
        }
    else:
        headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer {}".format(encoded_jwt)
        }

    # Get all available services
    url = app_settings.BACKEND_SERVICES_URL

    response = requests.get(url=url, headers=headers, timeout=10, verify=False)
    response.raise_for_status()

    sso_services = response.json()

    # Find out the id for current service
    service_id = None
    for sso_service in sso_services:
        if sso_service["service_url"] == app_settings.APP_URL:
            service_id = sso_service['id']
            break
    if not service_id:
        _msg = ("Current service {} is not listed in SSO!"
                .format(app_settings.APP_URL))
        raise Exception(_msg)

    # Subscribe to current service
    url = (app_settings.USER_SUBSCRIPTIONS_CREATE_URL.format(sso_id=sso_id, service_id=service_id))

    response = requests.post(url=url, headers=headers, timeout=10, verify=False)
    response.raise_for_status()

    logger.info("Profile with SSO ID {sso_id} was successfully subscribed to"
                " service {service_name}!"
                .format(sso_id=sso_id,
                        service_name=app_settings.APP_URL))



import pytest

from django_sso_app.backend.users.models import User
from django_sso_app.backend.users.tests.factories import UserFactory


@pytest.fixture(autouse=True)
def media_storage(settings, tmpdir):
    settings.MEDIA_ROOT = tmpdir.strpath


@pytest.fixture
def user() -> User:
    return UserFactory()



import logging

from django.dispatch import receiver
from django.db.models.signals import post_save
from django.contrib.auth import get_user_model

from ..utils import get_or_create_user_profile, is_django_staff

logger = logging.getLogger('django_sso_app.core.apps.profiles.signals')

User = get_user_model()

LAST_LOGIN_UPDATED_FIELDS_OBJECT = frozenset({'last_login'})


# user

@receiver(post_save, sender=User)
def create_update_user_profile(sender, instance, created, raw, using, update_fields, **kwargs):
    """
    Creates new user profile if not already done.
    """
    user = instance

    if raw:
        # https://github.com/django/django/commit/18a2fb19074ce6789639b62710c279a711dabf97
        # setting property flag
        setattr(user, '__dssoa__loaddata', True)
        return

    if created:
        logger.debug('backend signals user created')

        setattr(user, '__dssoa__creating', True)

        profile = get_or_create_user_profile(user, commit=True)

        logger.debug('new profile created "{}"'.format(profile))

        # refreshing user instance property
        #user.sso_app_profile = profile

        setattr(user, '__dssoa__creating', False)

    else:
        logger.debug('backend signals user updated')
        profile = get_or_create_user_profile(user, commit=False)

        # rev should update only for non django staff users
        update_rev = not is_django_staff(user)

        # replaced by '__user_loggin_in'
        # if user.previous_serialized_as_string == user.serialized_as_string:
        if getattr(user, '__dssoa__creating', False):
            logger.info('user _creating')
            update_rev = False
        elif getattr(user, '__dssoa__updating_default_groups', False):
            logger.info('user updating default groups in')
            update_rev = False
        elif user.password_has_been_hardened:
            logger.info('password_has_been_hardened')
            update_rev = False
        elif is_django_staff(user):
            logger.info('User is staff, skipping rev update')
            update_rev = False
        elif update_fields == LAST_LOGIN_UPDATED_FIELDS_OBJECT:
            update_rev = False

        # align profile user fields
        if user.email_has_been_updated:
            # aligning sso_app_profile django_user_email
            profile.django_user_email = user.sso_app_new_email

        if user.username_has_been_updated:
            # aligning sso_app_profile django_user_username
            profile.django_user_username = user.sso_app_new_username

        if update_rev:
            logger.info('Update rev by User signal')
            profile.update_rev(True)  # updating rev
